#! /usr/bin/env python
"""Exclude records based on language as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass

import timeout_decorator
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.language_service
import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class ExcludeLanguagesPrep(JsonSchemaMixin):
    """Prepares records by excluding ones that are not in the languages_to_include"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    def __init__(self, *, prep_operation: colrev.ops.prep.Prep, settings: dict) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        # Note : the following objects have heavy memory footprints and should be
        # class (not object) properties to keep parallel processing as
        # efficient as possible (the object is passed to each thread)
        languages_to_include = ["eng"]
        # if not prep_operation.review_manager.in_ci_environment():
        self.language_service = colrev.env.language_service.LanguageService()

        prescreen_package_endpoints = (
            prep_operation.review_manager.settings.prescreen.prescreen_package_endpoints
        )
        if "scope_prescreen" in [s["endpoint"] for s in prescreen_package_endpoints]:
            for scope_prescreen in [
                s
                for s in prescreen_package_endpoints
                if "scope_prescreen" == s["endpoint"]
            ]:
                languages_to_include.extend(
                    scope_prescreen.get("LanguageScope", ["eng"])
                )
        self.language_service.validate_iso_639_3_language_codes(
            lang_code_list=languages_to_include
        )
        self.languages_to_include = list(set(languages_to_include))

    @timeout_decorator.timeout(60, use_signals=False)
    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare the record by excluding records whose metadata is not in English"""

        # pylint: disable=too-many-statements
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-return-statements

        # Note : other languages are not yet supported
        # because the dedupe does not yet support cross-language merges

        if "language" in record.data:
            if record.data["language"] not in self.languages_to_include:
                record.prescreen_exclude(
                    reason=(
                        "language of title not in "
                        f"[{','.join(self.languages_to_include)}]"
                    )
                )
            return record

        # To avoid misclassifications for short titles
        if len(record.data.get("title", "")) < 30:
            # If language not in record, add language
            # (always - needed in dedupe.)
            record.data["language"] = "eng"
            return record

        # Deal with title fields containing titles in two languages
        if (
            len(record.data.get("title", "")) > 40
            and record.data.get("title", "").count("[") == 1
        ):
            confidence_values_part1 = (
                self.language_service.compute_language_confidence_values(
                    text=record.data["title"].split("[")[0]
                )
            )
            confidence_values_part2 = (
                self.language_service.compute_language_confidence_values(
                    text=record.data["title"].split("[")[1]
                )
            )

            if len(confidence_values_part1) == 0 or len(confidence_values_part2) == 0:
                record.set_status(
                    target_state=colrev.record.RecordState.md_needs_manual_preparation
                )
                return record

            lang_1, conf_1 = confidence_values_part1[0]
            lang_2, conf_2 = confidence_values_part2[0]

            if conf_1 < 0.8 and conf_2 < 0.8:
                record.update_field(
                    key="title",
                    value=record.data.get("title", ""),
                    source="",
                    note="quality_defect,language-not-found",
                    append_edit=True,
                )
                record.remove_field(key="language")
                record.set_status(
                    target_state=colrev.record.RecordState.md_needs_manual_preparation
                )
                return record

            if "eng" == lang_1:
                record.update_field(
                    key=f"title_{lang_2}",
                    value=record.data["title"].split("[")[1].rstrip("]"),
                    source="LanguageDetector_split",
                )
                record.update_field(
                    key="title",
                    value=record.data["title"].split("[")[0].rstrip(),
                    source="LanguageDetector_split",
                )
                record.update_field(
                    key="language",
                    value="eng",
                    source="LanguageDetector",
                    note="",
                )
            else:
                record.update_field(
                    key=f"title_{lang_1}",
                    value=record.data["title"].split("[")[0].rstrip(),
                    source="LanguageDetector_split",
                )
                record.update_field(
                    key="title",
                    value=record.data["title"].split("[")[1].rstrip(),
                    source="LanguageDetector_split",
                )
                record.update_field(
                    key="language",
                    value="eng",
                    source="LanguageDetector",
                    note="",
                )
                record.prescreen_exclude(
                    reason=f"language of title(s) not in [{','.join(self.languages_to_include)}]"
                )

            return record

        confidence_values = self.language_service.compute_language_confidence_values(
            text=record.data["title"]
        )

        if len(confidence_values) == 0:
            record.update_field(
                key="title",
                value=record.data.get("title", ""),
                source="LanguageDetector",
                note="cannot_predict_language",
            )
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )
            return record

        predicted_language, conf = confidence_values.pop()

        if conf > 0.8:
            record.update_field(
                key="language",
                value=predicted_language,
                source="LanguageDetector",
                note="",
                append_edit=False,
            )

        else:
            record.update_field(
                key="title",
                value=record.data.get("title", ""),
                source="",
                note="quality_defect,language-not-found",
                append_edit=True,
            )
            record.remove_field(key="language")
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )

        if record.data.get("language", "") not in self.languages_to_include:
            record.prescreen_exclude(
                reason=f"language of title not in [{','.join(self.languages_to_include)}]"
            )

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Removal of broken IDs as a prep operation"""
from __future__ import annotations

import re
from dataclasses import dataclass

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search_sources
import colrev.record

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class RemoveBrokenIDPrep(JsonSchemaMixin):
    """Prepares records by removing invalid IDs DOIs/ISBNs"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    # check_status: relies on crossref / openlibrary connectors!
    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare the record by removing broken IDs (invalid DOIs/ISBNs)"""

        if prep_operation.polish and not prep_operation.force_mode:
            return record

        if "doi" in record.data:
            # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
            doi_match = re.match(r"^10.\d{4,9}\/", record.data["doi"])
            if not doi_match:
                record.remove_field(key="doi")
        if "isbn" in record.data:
            try:
                session = prep_operation.review_manager.get_cached_session()

                isbn = record.data["isbn"].replace("-", "").replace(" ", "")
                url = f"https://openlibrary.org/isbn/{isbn}.json"
                ret = session.request(
                    "GET",
                    url,
                    headers=prep_operation.requests_headers,
                    timeout=prep_operation.timeout,
                )
                ret.raise_for_status()
            except requests.exceptions.RequestException:
                record.remove_field(key="isbn")
        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""SearchSource: Springer Link"""
from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code

# Note : API requires registration
# https://dev.springernature.com/


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class SpringerLinkSearchSource(JsonSchemaMixin):
    """SearchSource for Springer Link"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "Springer Link"
    link = "https://link.springer.com/"

    def __init__(
        self, *, source_operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

        search_operation.review_manager.logger.debug(
            f"Validate SearchSource {source.filename}"
        )

        if "query_file" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                f"Source missing query_file search_parameter ({source.filename})"
            )

        if not Path(source.search_parameters["query_file"]).is_file():
            raise colrev_exceptions.InvalidQueryException(
                f"File does not exist: query_file {source.search_parameters['query_file']} "
                f"for ({source.filename})"
            )

        search_operation.review_manager.logger.debug(
            f"SearchSource {source.filename} validated"
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for Springer Link"""

        result = {"confidence": 0.1}

        if filename.suffix == ".csv":
            if data.count("http://link.springer.com") == data.count("\n"):
                result["confidence"] = 1.0
                return result

        # Note : no features in bib file for identification

        return result

    @classmethod
    def add_endpoint(
        cls, search_operation: colrev.ops.search.Search, query: str
    ) -> typing.Optional[colrev.settings.SearchSource]:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        return None

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of SpringerLink"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for Springer Link"""

        # pylint: disable=too-many-branches

        for record in records.values():
            if "item_title" in record:
                record["title"] = record["item_title"]
                del record["item_title"]

            if "content_type" in record:
                if "Article" == record["content_type"]:
                    if "publication_title" in record:
                        record["journal"] = record["publication_title"]
                        del record["publication_title"]
                    record.change_entrytype(new_entrytype="article")

                if "Book" == record["content_type"]:
                    if "publication_title" in record:
                        record["series"] = record["publication_title"]
                        del record["publication_title"]
                    record.change_entrytype(new_entrytype="book")

                if "Chapter" == record["content_type"]:
                    record["chapter"] = record["title"]
                    if "publication_title" in record:
                        record["title"] = record["publication_title"]
                        del record["publication_title"]
                    record.change_entrytype(new_entrytype="inbook")

                del record["content_type"]

            if "item_doi" in record:
                record["doi"] = record["item_doi"]
                del record["item_doi"]
            if "journal_volume" in record:
                record["volume"] = record["journal_volume"]
                del record["journal_volume"]
            if "journal_issue" in record:
                record["number"] = record["journal_issue"]
                del record["journal_issue"]

            # Fix authors
            if "author" in record:
                # a-bd-z: do not match McDonald
                record["author"] = re.sub(
                    r"([a-bd-z]{1})([A-Z]{1})", r"\g<1> and \g<2>", record["author"]
                )

        return records

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for Springer Link"""

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Checks of consistency between global IDs a prep operation"""
from __future__ import annotations

from dataclasses import dataclass

import timeout_decorator
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from thefuzz import fuzz

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.crossref as crossref_connector
import colrev.ops.built_in.search_sources.website as website_connector
import colrev.ops.search_sources
import colrev.record

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class GlobalIDConsistencyPrep(JsonSchemaMixin):
    """Prepares records by removing IDs (DOIs/URLs) that do not match with the metadata"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = "check with the developer"
    always_apply_changes = True

    __fields_to_check = ["author", "title", "journal", "year", "volume", "number"]

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        self.url_connector = website_connector.WebsiteConnector(
            source_operation=prep_operation
        )
        self.prep_operation = prep_operation

    def __validate_against_doi_metadata(self, *, record: colrev.record.Record) -> None:
        # pylint: disable=too-many-branches

        if "doi" not in record.data:
            return
        if "doi" in record.data["colrev_data_provenance"]:
            if (
                "md_curated.bib"
                in record.data["colrev_data_provenance"]["doi"]["source"]
            ):
                return

        record_copy = record.copy_prep_rec()

        crossref_search_source = crossref_connector.CrossrefSearchSource(
            source_operation=self.prep_operation
        )
        crossref_md = crossref_search_source.get_masterdata(
            prep_operation=self.prep_operation, record=record_copy, save_feed=False
        )

        for key, value in crossref_md.data.items():
            if key not in self.__fields_to_check:
                continue
            if not isinstance(value, str):
                continue
            if key in record.data:
                if "UNKNOWN" == record.data[key]:
                    continue
                if key in ["author", "title", "journal"]:
                    if len(crossref_md.data[key]) < 5 or len(record.data[key]) < 5:
                        continue
                if (
                    fuzz.partial_ratio(
                        record.data[key].lower(), crossref_md.data[key].lower()
                    )
                    < 70
                ):
                    if record.masterdata_is_curated():
                        record.remove_field(key="doi")
                    else:
                        record.data[
                            "colrev_status"
                        ] = colrev.record.RecordState.md_needs_manual_preparation
                        record.add_masterdata_provenance_note(
                            key=key, note="disagreement with doi metadata"
                        )

    def __validate_against_url_metadata(self, *, record: colrev.record.Record) -> None:
        if "url" not in record.data:
            return

        if any(x in record.data["url"] for x in ["search.ebscohost.com/login"]):
            return

        if "md_curated.bib" in record.data["colrev_data_provenance"]["url"]["source"]:
            return

        try:
            url_record = record.copy_prep_rec()
            self.url_connector.retrieve_md_from_website(
                record=url_record, prep_operation=self.prep_operation
            )
            for key, value in url_record.data.items():
                if key not in self.__fields_to_check:
                    continue
                if not isinstance(value, str):
                    continue
                if key in record.data:
                    if len(url_record.data[key]) < 5 or len(record.data[key]) < 5:
                        continue
                    if (
                        fuzz.partial_ratio(
                            record.data[key].lower(), url_record.data[key].lower()
                        )
                        < 70
                    ):
                        if record.masterdata_is_curated():
                            record.remove_field(key="url")
                        else:
                            record.data[
                                "colrev_status"
                            ] = colrev.record.RecordState.md_needs_manual_preparation
                            record.add_masterdata_provenance_note(
                                key=key,
                                note=f"disagreement with website metadata ({value})",
                            )
        except AttributeError:
            pass

    @timeout_decorator.timeout(40, use_signals=False)
    def prepare(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.PrepRecord,
    ) -> colrev.record.Record:
        """Prepare records by removing IDs (DOIs/URLs) that do not match with the metadata

        When metadata provided by DOI/crossref or on the website (url) differs from
        the RECORD: set status to md_needs_manual_preparation."""

        if prep_operation.polish and not prep_operation.force_mode:
            return record

        self.__validate_against_doi_metadata(record=record)

        self.__validate_against_url_metadata(record=record)

        return record


if __name__ == "__main__":
    pass
